#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

""" Defines the dispatcher exception. """

__author__ = 'Vincenzo Arcidiacono <vinci1it2000@gmail.com>'


class DispatcherError(Exception):

    def __reduce__(self):
        fn, args, state = super(DispatcherError, self).__reduce__()
        state = {k: v for k, v in state.items() if k not in ('sol', 'plot')}
        return fn, args, state

    def __init__(self, *args, sol=None, ex=None, **kwargs):
        # noinspection PyArgumentList
        super(DispatcherError, self).__init__(*args, **kwargs)
        self.plot = None
        self.sol = None
        self.ex = ex
        self.update(sol)

    def update(self, sol):
        self.sol = sol
        self.plot = self.sol.plot if sol is not None else None


class DispatcherAbort(BaseException):
    pass


class SkipNode(BaseException):
    def __init__(self, *args, ex=None, **kwargs):
        # noinspection PyArgumentList
        super(SkipNode, self).__init__(*args, **kwargs)
        self.ex = ex


class ExecutorShutdown(BaseException):
    pass


class WebResponse(BaseException):
    def __init__(self, response):
        self.response = response



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Dispatcher directive.
"""

import glob
import sphinx
import shutil
import posixpath
import os.path as osp
import schedula as sh
from docutils import nodes
# noinspection PyPep8Naming
from docutils.parsers.rst import directives
from sphinx.ext.autodoc import bool_option
from sphinx.ext.graphviz import (
    Graphviz, latex_visit_graphviz, texinfo_visit_graphviz, text_visit_graphviz,
    man_visit_graphviz
)
from sphinx.util.i18n import search_image_for_language


class dsp(nodes.General, nodes.Inline, nodes.Element):
    pass


class DispatcherSphinxDirective(Graphviz):
    required_arguments = 1
    img_opt = {
        'height': directives.length_or_unitless,
        'width': directives.length_or_percentage_or_unitless,
    }
    option_spec = {
        'graphviz_dot': directives.unchanged,  # sphinx==1.3.5
        'index': bool_option,
        'viz': bool_option
    }
    sh.combine_dicts(img_opt, Graphviz.option_spec, base=option_spec)

    def run(self):
        node = super(DispatcherSphinxDirective, self).run()[0]
        # noinspection PyUnresolvedReferences
        node = dsp(node.rawsource, *node.children, **node.attributes)
        node['img_opt'] = sh.selector(
            self.img_opt, self.options, allow_miss=True
        )
        node['index'] = self.options.get('index', False)
        env = self.state.document.settings.env
        argument = search_image_for_language(self.arguments[0], env)
        dirpath = osp.dirname(env.relfn2path(argument)[1])
        node['dirpath'] = dirpath if osp.isdir(dirpath) else None
        return [node]


def html_visit_dispatcher(self, node):
    dirpath = node['dirpath']
    dname = osp.basename(dirpath)
    outd = posixpath.join(self.builder.outdir, self.builder.imagedir, dname)
    if not osp.isdir(outd):
        shutil.copytree(dirpath, outd)
    if node['index']:
        fname = 'index.html'
    else:
        fname = osp.basename(glob.glob(osp.join(outd, '*.gv'))[0])
        fname = '%s.html?controls=true' % osp.splitext(fname)[0]
    attr = sh.combine_dicts(node.get('img_opt', {}), base=dict(
        src=posixpath.join(self.builder.imgpath, dname, fname),
        width='100%', height='500px', frameborder='0'
    ))
    self.body.append('<iframe %s allowfullscreen></iframe>' % ' '.join(
        '%s="%s"' % v for v in attr.items()
    ))
    raise nodes.SkipNode


def setup(app):
    """Setup `dsp` Sphinx extension module. """
    app.setup_extension('sphinx.ext.graphviz')  # To set all defaults.
    app.add_node(
        dsp,
        html=(html_visit_dispatcher, None),
        latex=(latex_visit_graphviz, None),
        texinfo=(texinfo_visit_graphviz, None),
        text=(text_visit_graphviz, None),
        man=(man_visit_graphviz, None)
    )
    app.add_directive('dsp', DispatcherSphinxDirective)
    return {'version': sphinx.__display_version__, 'parallel_read_safe': True}



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
import os
import unittest


EXTRAS = os.environ.get('EXTRAS', 'all')


@unittest.skipIf(os.name == 'nt', 'Not for os %s.' % os.name)
@unittest.skipIf(EXTRAS not in ('all',), 'Not for extra %s.' % EXTRAS)
class TestDoctest(unittest.TestCase):
    def runTest(self):
        import doctest
        failure_count, test_count = doctest.testfile(
            '../README.rst',
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
        )
        self.assertGreater(test_count, 0, (failure_count, test_count))
        self.assertEqual(failure_count, 0, (failure_count, test_count))



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build a form flask app from a dispatcher.
"""

import datetime
import os.path as osp
import schedula as sh
from .mail import Mail
from .config import Config
from flask_sqlalchemy import SQLAlchemy
from flask_babel import Babel, lazy_gettext
from werkzeug.datastructures import MultiDict
from flask_wtf.recaptcha import RecaptchaField
from flask_principal import Permission, RoleNeed
from flask_security.models import fsqla_v3 as fsqla
from flask import after_this_request, request, jsonify, current_app, flash
from sqlalchemy import Column, String, Integer, DateTime, ForeignKey, JSON
from flask_security.utils import (
    base_render_json, suppress_form_csrf, view_commit
)
from flask_security.forms import (
    ConfirmRegisterForm, Required, StringField, Form, EmailField
)
from flask_security import (
    Security, SQLAlchemyUserDatastore, current_user as cu, auth_required
)


def default_get_form_context():
    return {
        'userInfo': getattr(cu, "get_security_payload", lambda: {})(),
        'reCAPTCHA': current_app.config.get('RECAPTCHA_PUBLIC_KEY')
    }


def basic_app(sitemap, app):
    app.config.from_object(Config)
    if getattr(sitemap, 'basic_app_config'):
        app.config.from_object(sitemap.basic_app_config)

    # Create database connection object
    db = SQLAlchemy(app)

    def default_name(context):
        return f'Item {context.get_current_parameters()["id"]}'

    def is_admin():
        return Permission(RoleNeed('admin')).can()

    class Item(db.Model):
        __tablename__ = 'item'
        id = Column(Integer, primary_key=True)
        name = Column(String(255), default=default_name)
        category = Column(String(255))
        data = Column('data', JSON)
        user_id = Column(Integer, ForeignKey('user.id'))
        created_at = Column(DateTime(), default=datetime.datetime.utcnow)
        updated_at = Column(DateTime(), onupdate=datetime.datetime.utcnow)

        def payload(self, data=False):
            res = {
                'id': self.id,
                'name': self.name,
                'created_at': self.created_at,
                'updated_at': self.updated_at
            }
            for k in ('created_at', 'updated_at'):
                if res[k]:
                    res[k] = res[k].isoformat()

            if data:
                res['data'] = self.data
            return res

        def __repr__(self):
            return f'{self.category}-{self.id}-{self.user_id}'

    @app.route('/item/<category>', methods=['GET', 'POST'])
    @app.route('/item/<category>/<int:id_item>', methods=[
        'GET', 'PUT', 'PATCH', 'DELETE'
    ])
    @auth_required()
    def item(category, id_item=None):
        args = request.args
        method = request.method
        is_get = method == 'GET'
        kw = {'category': category, 'user_id': cu.id}
        if not is_get:
            kw['data'] = request.get_json()
        if 'name' in args:
            kw['name'] = args.get("name", type=str)

        by = {'category': category, 'user_id': cu.id}
        if id_item is not None:
            by['id'] = kw['id'] = id_item
        if is_admin():
            by.pop('user_id')
        if method == 'POST':  # Create.
            item = Item(**kw)
            db.session.add(item)
            db.session.flush()
            payload = item.payload()
        else:  # Read, Delete, Update/Modify, Update/Replace.
            query = Item.query.filter_by(**by)
            if id_item is None:  # GET
                query = query.order_by(Item.id)
                if 'page' in args and 'per_page' in args:
                    pag = db.paginate(
                        query,
                        page=args.get("page", type=int),
                        max_per_page=args.get("per_page", type=int),
                        count=True, error_out=False
                    )
                    items = [item.payload() for item in pag.items]
                    payload = {'page': pag.page, 'items': items,
                               'total': pag.total}
                else:
                    items = [item.payload() for item in query.all()]
                    payload = {'items': items, 'total': len(items)}
            else:
                item = query.first()
                if method == 'DELETE':
                    db.session.delete(item)
                elif method in ('PATCH', 'PUT'):
                    if method == 'PATCH':
                        kw['data'] = sh.combine_nested_dicts(
                            item.data, kw['data']
                        )
                    for k, v in kw.items():
                        setattr(item, k, v)
                    db.session.add(item)
                    db.session.flush()
                payload = item.payload(data=is_get)
        is_get or after_this_request(view_commit)
        return jsonify(payload)

    sitemap.add2csrf_protected(item=('view', 'item'))

    # Define models
    fsqla.FsModels.set_db_info(db)

    class Role(db.Model, fsqla.FsRoleMixin):
        pass

    class User(db.Model, fsqla.FsUserMixin):
        firstname = Column(String(255))
        lastname = Column(String(255))

        def get_security_payload(self):
            return {k: v for k, v in {
                'email': self.email,
                'username': self.username,
                'firstname': self.firstname,
                'lastname': self.lastname
            }.items() if v is not None}

    # Setup Flask-Security
    class EditForm(Form):
        firstname = StringField('firstname', [Required()])
        lastname = StringField('lastname', [Required()])

    class ExtendedConfirmRegisterForm(ConfirmRegisterForm, EditForm):
        pass

    user_datastore = SQLAlchemyUserDatastore(db, User, Role)
    app.security = Security(
        app, user_datastore, confirm_register_form=ExtendedConfirmRegisterForm
    )
    sitemap.add2csrf_protected(item=('bp', app.security.blueprint_name))

    @app.route('/locales/<language>/<namespace>', methods=['GET'])
    def locales(language, namespace):
        from werkzeug.exceptions import NotFound
        from flask import send_from_directory

        for d in app.config['SCHEDULA_I18N_DIRNAME']:
            try:
                return send_from_directory(
                    d, f'{osp.join(language, "LC_MESSAGES", namespace)}.po',
                    as_attachment=True
                )
            except NotFound:
                pass
        raise NotFound()

    @app.route(f'{app.config["SECURITY_URL_PREFIX"]}/edit', methods=['POST'])
    @auth_required()
    def edit():
        data = MultiDict(
            request.get_json()) if request.is_json else request.form
        form = EditForm(data, meta=suppress_form_csrf())
        form.user = cu
        if form.validate_on_submit():
            after_this_request(view_commit)
            for k, v in form.data.items():
                setattr(cu, k, v)
            db.session.add(cu)
        return base_render_json(form)

    sitemap.add2csrf_protected(item=('view', 'edit'))

    class ContactForm(Form):
        name = StringField('name', [Required()])
        email = EmailField('email', [Required()])
        subject = StringField('subject', [Required()])
        message = StringField('message', [Required()])
        recaptcha = RecaptchaField('g-recaptcha-response')

    @app.route('/mail/contact', methods=['POST'])
    def contact():
        data = MultiDict(
            request.get_json()
        ) if request.is_json else request.form
        if cu.is_authenticated:
            if 'email' not in data:
                data['email'] = cu.email
            if 'name' not in data:
                data['name'] = f'{cu.firstname} {cu.lastname}'
        form = ContactForm(data, meta=suppress_form_csrf())
        if form.validate_on_submit():
            mail.send_rst(
                to=[form.data['email'],
                    current_app.config.get('MAIL_DEFAULT_SENDER')],
                rst='contact', reply_to=form.data['email'], user=cu, data=data,
                created=datetime.datetime.now().strftime("%d/%m/%Y-%H:%M:%S")
            )
            flash(
                str(lazy_gettext('Your message has been successfully sent!')),
                'success'
            )
        return base_render_json(form)

    def get_locale():
        from flask import request
        return request.accept_languages.best_match(['it_IT', 'en_US'])

    Babel(app, default_locale='en', locale_selector=get_locale)
    mail = Mail(app)
    return app



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It is a patch to sphinx.ext.autosummary.
"""
import warnings
import logging
import os.path as osp
from typing import Any
from sphinx.util.inspect import getall
from sphinx.util.osutil import ensuredir
from sphinx.ext.autosummary import (
    import_by_name, get_documenter, get_rst_suffix, mock, import_ivar_by_name,
    ImportExceptionGroup
)
from sphinx.ext.autosummary.generate import (
    find_autosummary_in_files, AutosummaryRenderer, ModuleScanner, _get_members,
    _get_module_attrs, _get_modules, _split_full_qualified_name
)
from sphinx.locale import __

logger = logging.getLogger(__name__)
warnings.filterwarnings(
    'ignore', category=DeprecationWarning, module='docutils'
)


def generate_autosummary_content(
        name, obj, parent, template, template_name, imported_members, app,
        recursive, context, modname=None, qualname=None):
    doc = get_documenter(app, obj, parent)

    ns: dict[str, Any] = {}
    ns.update(context)

    if doc.objtype == 'module':
        scanner = ModuleScanner(app, obj)
        ns['members'] = scanner.scan(imported_members)

        respect_module_all = not app.config.autosummary_ignore_module_all
        imported_members = imported_members or (
                '__all__' in dir(obj) and respect_module_all)

        ns['functions'], ns['all_functions'] = \
            _get_members(doc, app, obj, {'function'}, imported=imported_members)
        ns['functions'], ns['all_functions'] = \
            _get_members(doc, app, obj, {'function'}, imported=imported_members)
        ns['classes'], ns['all_classes'] = \
            _get_members(doc, app, obj, {'class'}, imported=imported_members)
        ns['exceptions'], ns['all_exceptions'] = \
            _get_members(doc, app, obj, {'exception'},
                         imported=imported_members)
        ns['attributes'], ns['all_attributes'] = \
            _get_module_attrs(name, ns['members'])
        ns['dispatchers'], ns['all_dispatchers'] = \
            _get_members(doc, app, obj, {'dispatcher'},
                         imported=imported_members)
        ispackage = hasattr(obj, '__path__')
        if ispackage and recursive:
            # Use members that are not modules as skip list, because it would then mean
            # that module was overwritten in the package namespace
            skip = (
                    ns["all_functions"]
                    + ns["all_classes"]
                    + ns["all_exceptions"]
                    + ns["all_attributes"]
                    + ns["all_dispatchers"]
            )

            # If respect_module_all and module has a __all__ attribute, first get
            # modules that were explicitly imported. Next, find the rest with the
            # get_modules method, but only put in "public" modules that are in the
            # __all__ list
            #
            # Otherwise, use get_modules method normally
            if respect_module_all and '__all__' in dir(obj):
                imported_modules, all_imported_modules = \
                    _get_members(doc, app, obj, {'module'}, imported=True)
                skip += all_imported_modules
                imported_modules = [name + '.' + modname for modname in
                                    imported_modules]
                all_imported_modules = \
                    [name + '.' + modname for modname in all_imported_modules]
                public_members = getall(obj)
            else:
                imported_modules, all_imported_modules = [], []
                public_members = None

            modules, all_modules = _get_modules(obj, skip=skip, name=name,
                                                public_members=public_members)
            ns['modules'] = imported_modules + modules
            ns["all_modules"] = all_imported_modules + all_modules
    elif doc.objtype == 'class':
        ns['members'] = dir(obj)
        ns['inherited_members'] = \
            set(dir(obj)) - set(obj.__dict__.keys())
        ns['methods'], ns['all_methods'] = \
            _get_members(doc, app, obj, {'method'}, include_public={'__init__'})
        ns['attributes'], ns['all_attributes'] = \
            _get_members(doc, app, obj, {'attribute', 'property'})

    if modname is None or qualname is None:
        modname, qualname = _split_full_qualified_name(name)

    if doc.objtype in ('method', 'attribute', 'property'):
        ns['class'] = qualname.rsplit(".", 1)[0]

    if doc.objtype in ('class',):
        shortname = qualname
    else:
        shortname = qualname.rsplit(".", 1)[-1]

    ns['fullname'] = name
    ns['module'] = modname
    ns['objname'] = qualname
    ns['name'] = shortname

    ns['objtype'] = doc.objtype
    ns['underline'] = len(name) * '='

    if template_name:
        return template.render(template_name, ns)
    else:
        return template.render(doc.objtype, ns)


def generate_autosummary_docs(
        sources, output_dir=None, suffix='.rst', base_path=None,
        imported_members=False, app=None, overwrite=True, encoding='utf-8'):
    showed_sources = sorted(sources)
    if len(showed_sources) > 20:
        showed_sources = showed_sources[:10] + ['...'] + showed_sources[-10:]
    logger.info(__('[autosummary] generating autosummary for: %s') %
                ', '.join(showed_sources))

    if output_dir:
        logger.info(__('[autosummary] writing to %s') % output_dir)

    if base_path is not None:
        sources = [osp.join(base_path, filename) for filename in sources]

    template = AutosummaryRenderer(app)

    # read
    items = find_autosummary_in_files(sources)

    # keep track of new files
    new_files = []

    if app:
        filename_map = app.config.autosummary_filename_map
    else:
        filename_map = {}

    # write
    for entry in sorted(set(items), key=str):
        if entry.path is None:
            # The corresponding autosummary:: directive did not have
            # a :toctree: option
            continue

        path = output_dir or osp.abspath(entry.path)
        ensuredir(path)

        try:
            name, obj, parent, modname = import_by_name(entry.name)
            qualname = name.replace(modname + ".", "")
        except ImportExceptionGroup as exc:
            try:
                # try to import as an instance attribute
                name, obj, parent, modname = import_ivar_by_name(entry.name)
                qualname = name.replace(modname + ".", "")
            except ImportError as exc2:
                if exc2.__cause__:
                    exceptions: list[BaseException] = exc.exceptions + [
                        exc2.__cause__]
                else:
                    exceptions = exc.exceptions + [exc2]

                errors = list(
                    {f"* {type(e).__name__}: {e}" for e in exceptions})
                logger.warning(
                    __('[autosummary] failed to import %s.\nPossible hints:\n%s'),
                    entry.name, '\n'.join(errors))
                continue

        context: dict[str, Any] = {}
        if app:
            context.update(app.config.autosummary_context)

        content = generate_autosummary_content(name, obj, parent, template,
                                               entry.template,
                                               imported_members, app,
                                               entry.recursive, context,
                                               modname, qualname)

        filename = osp.join(path, filename_map.get(name, name) + suffix)
        if osp.isfile(filename):
            with open(filename, encoding=encoding) as f:
                old_content = f.read()

            if content == old_content:
                continue
            if overwrite:  # content has changed
                with open(filename, 'w', encoding=encoding) as f:
                    f.write(content)
                new_files.append(filename)
        else:
            with open(filename, 'w', encoding=encoding) as f:
                f.write(content)
            new_files.append(filename)

    # descend recursively to new files
    if new_files:
        generate_autosummary_docs(new_files, output_dir=output_dir,
                                  suffix=suffix, base_path=base_path,
                                  imported_members=imported_members, app=app,
                                  overwrite=overwrite)


def process_generate_options(app):
    genfiles = app.config.autosummary_generate

    if genfiles is True:
        env = app.builder.env
        genfiles = [env.doc2path(x, base=False) for x in env.found_docs
                    if osp.isfile(env.doc2path(x))]
    elif genfiles is False:
        pass
    else:
        ext = list(app.config.source_suffix)
        genfiles = [
            genfile + (ext[0] if not genfile.endswith(tuple(ext)) else '')
            for genfile in genfiles]

        for entry in genfiles[:]:
            if not osp.isfile(osp.join(app.srcdir, entry)):
                logger.warning(__('autosummary_generate: file not found: %s'),
                               entry)
                genfiles.remove(entry)

    if not genfiles:
        return

    suffix = get_rst_suffix(app)
    if suffix is None:
        logger.warning(__('autosummary generats .rst files internally. '
                          'But your source_suffix does not contain .rst. Skipped.'))
        return

    imported_members = app.config.autosummary_imported_members
    with mock(app.config.autosummary_mock_imports):
        generate_autosummary_docs(genfiles, suffix=suffix, base_path=app.srcdir,
                                  app=app, imported_members=imported_members,
                                  overwrite=app.config.autosummary_generate_overwrite,
                                  encoding=app.config.source_encoding)


def setup(app):
    app.setup_extension('sphinx.ext.autosummary')
    app.add_config_value('autosummary_skip_members', [], 'html')

    # replace callback process_generate_options of 'builder-inited' event.
    import sphinx.ext.autosummary as mdl
    pgo = mdl.process_generate_options
    event = 'builder-inited'
    listeners = app.events.listeners[event]
    from sphinx.events import EventListener
    for i, event in enumerate(listeners):
        if pgo in event:
            listeners[i] = EventListener(*(
                process_generate_options if e is pgo else e for e in event
            ))



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build a form flask app from a dispatcher.
"""
import os
import rst2txt
import schedula as sh
from flask import render_template
from flask_mail import Message, Mail as _Mail
from docutils.core import publish_string
from flask_babel import get_locale


def prepare_message(boby, subject, recipients, reply_to=None, **kwargs):
    body = publish_string(boby, writer=rst2txt.Writer()).decode()
    html = publish_string(boby, writer_name='html').decode()
    return Message(
        body=body, html=html, subject=subject, recipients=recipients,
        reply_to=reply_to, **kwargs
    )


def _render_template(*paths, **data):
    from jinja2.exceptions import TemplateNotFound
    for path in paths:
        try:
            return render_template(
                path, **data
            )
        except TemplateNotFound:
            continue
    raise TemplateNotFound


class Mail(_Mail):
    def send_rst(self, to, rst=None, reply_to=None, body=None, subject=None,
                 **data):
        from tabulate import tabulate
        language = get_locale().language
        body = body or _render_template(
            f'schedula/email/{rst}-body-{language}.rst',
            f'schedula/email/{rst}-body.rst',
            tabulate=tabulate, **data
        )
        subject = subject or _render_template(
            f'schedula/email/{rst}-subject-{language}.rst',
            f'schedula/email/{rst}-subject.rst',
            tabulate=tabulate, **data
        )
        recipients = list(sh.stlp(os.environ.get('LOCAL', to)))
        message = prepare_message(body, subject, recipients, reply_to=reply_to)
        return self.send(message)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides a base class for dispatcher objects.
"""

import copy
from .cst import NONE


class Base:
    """Base class for dispatcher objects."""

    def __new__(cls, *args, **kwargs):
        return super(Base, cls).__new__(cls)

    def __deepcopy__(self, memo):
        cls = self.__class__
        memo[id(self)] = result = cls.__new__(cls)
        for k, v in self.__dict__.items():
            # noinspection PyArgumentList
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    def web(self, depth=-1, node_data=NONE, node_function=NONE, directory=None,
            sites=None, run=True, subsite_idle_timeout=600):
        """
        Creates a dispatcher Flask app.

        :param depth:
            Depth of sub-dispatch API. If negative all levels are configured.
        :type depth: int, optional

        :param node_data:
            Data node attributes to produce API.
        :type node_data: tuple[str], optional

        :param node_function:
            Function node attributes produce API.
        :type node_function: tuple[str], optional

        :param directory:
            Where is the generated Flask app root located?
        :type directory: str, optional

        :param sites:
            A set of :class:`~schedula.utils.drw.Site` to maintain alive the
            backend server.
        :type sites: set[~schedula.utils.drw.Site], optional

        :param run:
            Run the backend server?
        :type run: bool, optional

        :param subsite_idle_timeout:
            Idle timeout of a debug subsite in seconds.
        :type subsite_idle_timeout: int, optional

        :return:
            A WebMap.
        :rtype: ~schedula.utils.web.WebMap

        Example:

        From a dispatcher like this:

        .. dispatcher:: dsp
           :opt: graph_attr={'ratio': '1'}
           :code:

            >>> from schedula import Dispatcher
            >>> dsp = Dispatcher(name='Dispatcher')
            >>> def fun(a):
            ...     return a + 1, a - 1
            >>> dsp.add_function('fun', fun, ['a'], ['b', 'c'])
            'fun'

        You can create a web server with the following steps::

            >>> print("Starting...\\n"); site = dsp.web(); site
            Starting...
            Site(WebMap([(Dispatcher, WebMap())]), host='localhost', ...)
            >>> import requests
            >>> url = '%s/%s/%s' % (site.url, dsp.name, fun.__name__)
            >>> requests.post(url, json={'args': (0,)}).json()['return']
            [1, -1]
            >>> site.shutdown()  # Remember to shutdown the server.
            True

        .. note::
           When :class:`~schedula.utils.drw.Site` is garbage collected, the
           server is shutdown automatically.
        """
        options = {'node_data': node_data, 'node_function': node_function}
        options = {k: v for k, v in options.items() if v is not NONE}
        from .web import WebMap
        from .sol import Solution

        obj = self.dsp if isinstance(self, Solution) else self

        webmap = WebMap()
        webmap.add_items(obj, workflow=False, depth=depth, **options)
        webmap.directory = directory
        webmap.idle_timeout = subsite_idle_timeout
        if sites is not None:
            sites.add(webmap.site(view=run))
        elif run:
            return webmap.site(view=run)
        return webmap

    def form(self, depth=1, node_data=NONE, node_function=NONE, directory=None,
             sites=None, run=True, view=True, get_context=NONE, get_data=NONE,
             edit_on_change=NONE, pre_submit=NONE, post_submit=NONE,
             subsite_idle_timeout=600, basic_app_config=None):
        """
        Creates a dispatcher Form Flask app.

        :param depth:
            Depth of sub-dispatch API. If negative all levels are configured.
        :type depth: int, optional

        :param node_data:
            Data node attributes to produce API.
        :type node_data: tuple[str], optional

        :param node_function:
            Function node attributes produce API.
        :type node_function: tuple[str], optional

        :param directory:
            Where is the generated Flask app root located?
        :type directory: str, optional

        :param sites:
            A set of :class:`~schedula.utils.drw.Site` to maintain alive the
            backend server.
        :type sites: set[~schedula.utils.drw.Site], optional

        :param run:
            Run the backend server?
        :type run: bool, optional

        :param view:
            Open the url site with the sys default opener.
        :type view: bool, optional

        :param get_context:
            Function to pass extra data as form context.
        :type get_context: function | dict, optional

        :param get_data:
            Function to initialize the formdata.
        :type get_data: function | dict, optional

        :param edit_on_change:
            Function to initialize the formdata.
        :type edit_on_change: function | dict, optional

        :param pre_submit:
            Function to initialize the formdata.
        :type pre_submit: function | dict, optional

        :param post_submit:
            Function to initialize the formdata.
        :type post_submit: function | dict, optional

        :param subsite_idle_timeout:
            Idle timeout of a debug subsite in seconds.
        :type subsite_idle_timeout: int, optional

        :param basic_app_config:
            Flask app config object.
        :type basic_app_config: object, optional

        :return:
            A FormMap or a Site if `sites is None` and `run or view is True`.
        :rtype: ~schedula.utils.form.FormMap | ~schedula.utils.drw.Site
        """
        options = {'node_data': node_data, 'node_function': node_function}
        options = {k: v for k, v in options.items() if v is not NONE}

        from .form import FormMap
        from .sol import Solution

        obj = self.dsp if isinstance(self, Solution) else self

        formmap = FormMap()
        formmap.add_items(obj, workflow=False, depth=depth, **options)
        formmap.directory = directory
        formmap.idle_timeout = subsite_idle_timeout
        formmap.basic_app_config = basic_app_config
        methods = {
            'get_form_context': get_context,
            'get_form_data': get_data,
            'get_edit_on_change_func': edit_on_change,
            'get_pre_submit_func': pre_submit,
            'get_post_submit_func': post_submit
        }
        for k, v in methods.items():
            if v is not NONE:
                setattr(formmap, f'_{k}', v)
        if sites is not None or run or view:
            site = formmap.site(view=view)
            site = run and not view and site.run() or site
            if sites is None:
                return site
            sites.add(site)
        return formmap

    def plot(self, workflow=None, view=True, depth=-1, name=NONE, comment=NONE,
             format=NONE, engine=NONE, encoding=NONE, graph_attr=NONE,
             node_attr=NONE, edge_attr=NONE, body=NONE, raw_body=NONE,
             node_styles=NONE, node_data=NONE, node_function=NONE,
             edge_data=NONE, max_lines=NONE, max_width=NONE, directory=None,
             sites=None, index=True, viz=False, short_name=None,
             executor='async', render=False, run=False):
        """
        Plots the Dispatcher with a graph in the DOT language with Graphviz.

        :param workflow:
           If True the latest solution will be plotted, otherwise the dmap.
        :type workflow: bool, optional

        :param view:
            Open the rendered directed graph in the DOT language with the sys
            default opener.
        :type view: bool, optional

        :param edge_data:
            Edge attributes to view.
        :type edge_data: tuple[str], optional

        :param node_data:
            Data node attributes to view.
        :type node_data: tuple[str], optional

        :param node_function:
            Function node attributes to view.
        :type node_function: tuple[str], optional

        :param node_styles:
            Default node styles according to graphviz node attributes.
        :type node_styles: dict[str|Token, dict[str, str]]

        :param depth:
            Depth of sub-dispatch plots. If negative all levels are plotted.
        :type depth: int, optional

        :param name:
            Graph name used in the source code.
        :type name: str

        :param comment:
            Comment added to the first line of the source.
        :type comment: str

        :param directory:
            (Sub)directory for source saving and rendering.
        :type directory: str, optional

        :param format:
            Rendering output format ('pdf', 'png', ...).
        :type format: str, optional

        :param engine:
            Layout command used ('dot', 'neato', ...).
        :type engine: str, optional

        :param encoding:
            Encoding for saving the source.
        :type encoding: str, optional

        :param graph_attr:
            Dict of (attribute, value) pairs for the graph.
        :type graph_attr: dict, optional

        :param node_attr:
            Dict of (attribute, value) pairs set for all nodes.
        :type node_attr: dict, optional

        :param edge_attr:
            Dict of (attribute, value) pairs set for all edges.
        :type edge_attr: dict, optional

        :param body:
            Dict of (attribute, value) pairs to add to the graph body.
        :type body: dict, optional

        :param raw_body:
            List of command to add to the graph body.
        :type raw_body: list, optional

        :param directory:
            Where is the generated Flask app root located?
        :type directory: str, optional

        :param sites:
            A set of :class:`~schedula.utils.drw.Site` to maintain alive the
            backend server.
        :type sites: set[~schedula.utils.drw.Site], optional

        :param index:
            Add the site index as first page?
        :type index: bool, optional

        :param max_lines:
            Maximum number of lines for rendering node attributes.
        :type max_lines: int, optional

        :param max_width:
            Maximum number of characters in a line to render node attributes.
        :type max_width: int, optional

        :param view:
            Open the main page of the site?
        :type view: bool, optional

        :param render:
            Render all pages statically?
        :type render: bool, optional

        :param viz:
            Use viz.js as back-end?
        :type viz: bool, optional

        :param short_name:
            Maximum length of the filename, if set name is hashed and reduced.
        :type short_name: int, optional

        :param executor:
            Pool executor to render object.
        :type executor: str, optional

        :return:
            A SiteMap or a Site if .
        :rtype: schedula.utils.drw.SiteMap

        Example:

        .. dispatcher:: dsp
           :opt: graph_attr={'ratio': '1'}
           :code:

            >>> from schedula import Dispatcher
            >>> dsp = Dispatcher(name='Dispatcher')
            >>> def fun(a):
            ...     return a + 1, a - 1
            >>> dsp.add_function('fun', fun, ['a'], ['b', 'c'])
            'fun'
            >>> dsp.plot(view=False, graph_attr={'ratio': '1'})
            SiteMap([(Dispatcher, SiteMap())])
        """

        d = {
            'name': name, 'comment': comment, 'format': format, 'body': body,
            'engine': engine, 'encoding': encoding, 'graph_attr': graph_attr,
            'node_attr': node_attr, 'edge_attr': edge_attr, 'raw_body': raw_body
        }
        options = {
            'digraph': {k: v for k, v in d.items() if v is not NONE} or NONE,
            'node_styles': node_styles,
            'node_data': node_data,
            'node_function': node_function,
            'edge_data': edge_data,
            'max_lines': max_lines,  # 5
            'max_width': max_width,  # 200
        }
        options = {k: v for k, v in options.items() if v is not NONE}
        from .drw import SiteMap
        from .sol import Solution

        if workflow is None and isinstance(self, Solution):
            workflow = True
        else:
            workflow = workflow or False

        sitemap = SiteMap()
        sitemap.short_name = short_name
        sitemap.directory = directory
        sitemap.add_items(self, workflow=workflow, depth=depth, **options)
        if render:
            sitemap.render(
                directory=directory, view=view, index=index, viz_js=viz,
                executor=executor
            )
        elif view or run or sites is not None:
            site = sitemap.site(
                directory, view=view, run=run, index=index, viz_js=viz,
                executor=executor
            )
            if sites is None:
                return site
            sites.add(site)
        return sitemap

    def get_node(self, *node_ids, node_attr=NONE):
        """
        Returns a sub node of a dispatcher.

        :param node_ids:
            A sequence of node ids or a single node id. The id order identifies
            a dispatcher sub-level.
        :type node_ids: str

        :param node_attr:
            Output node attr.

            If the searched node does not have this attribute, all its
            attributes are returned.

            When 'auto', returns the "default" attributes of the searched node,
            which are:

              - for data node: its output, and if not exists, all its
                attributes.
              - for function and sub-dispatcher nodes: the 'function' attribute.

            When 'description', returns the "description" of the searched node,
            searching also in function or sub-dispatcher input/output
            description.

            When 'output', returns the data node output.

            When 'default_value', returns the data node default value.

            When 'value_type', returns the data node value's type.

            When `None`, returns the node attributes.
        :type node_attr: str, None, optional

        :return:
            Node attributes and its real path.
        :rtype: (T, (str, ...))

        **Example**:

        .. dispatcher:: o
           :opt: graph_attr={'ratio': '1'}, depth=-1

            >>> import schedula as sh
            >>> sub_dsp = sh.Dispatcher(name='Sub-dispatcher')
            >>> def fun(a, b):
            ...     return a + b
            ...
            >>> sub_dsp.add_function('a + b', fun, ['a', 'b'], ['c'])
            'a + b'
            >>> dispatch = sh.SubDispatch(sub_dsp, ['c'], output_type='dict')
            >>> dsp = sh.Dispatcher(name='Dispatcher')
            >>> dsp.add_function('Sub-dispatcher', dispatch, ['a'], ['b'])
            'Sub-dispatcher'

            >>> o = dsp.dispatch(inputs={'a': {'a': 3, 'b': 1}})
            ...

        Get the sub node output::

            >>> dsp.get_node('Sub-dispatcher', 'c')
            (4, ('Sub-dispatcher', 'c'))
            >>> dsp.get_node('Sub-dispatcher', 'c', node_attr='type')
            ('data', ('Sub-dispatcher', 'c'))

        .. dispatcher:: sub_dsp
           :opt: workflow=True, graph_attr={'ratio': '1'}
           :code:

            >>> sub_dsp, sub_dsp_id = dsp.get_node('Sub-dispatcher')
        """
        kw = {}

        from .sol import Solution
        if node_attr is NONE:
            node_attr = 'output' if isinstance(self, Solution) else 'auto'

        if isinstance(self, Solution):
            kw['solution'] = self

        from .alg import get_sub_node
        dsp = getattr(self, 'dsp', self)

        # Returns the node.
        return get_sub_node(dsp, node_ids, node_attr=node_attr, **kw)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to plot dispatcher map and workflow.

Sub-Modules:

.. currentmodule:: schedula.utils.drw

.. autosummary::
    :nosignatures:
    :toctree: drw/

    nodes
"""

import os
import glob
import copy
import time
import html
import regex
import socket
import pprint
import string
import shutil
import inspect
import hashlib
import weakref
import logging
import datetime
import platform
import tempfile
import functools
import itertools
import threading
import collections
import os.path as osp
import urllib.parse as urlparse

try:
    from jinja2 import Environment, PackageLoader
    from pygments import highlight
    from pygments.lexers import Python3Lexer
    from pygments.formatters import HtmlFormatter
except ModuleNotFoundError:
    pass
from ..cst import START, SINK, END, EMPTY, SELF, NONE, PLOT
from ..dsp import (
    SubDispatch, combine_dicts, map_dict, combine_nested_dicts, selector, stlp,
    parent_func, get_nested_dicts, NoSub
)
from ..gen import counter
from ..asy import _async_executor, atexit_register, Future, _sync_executor
from ..asy.factory import ExecutorFactory

__author__ = 'Vincenzo Arcidiacono <vinci1it2000@gmail.com>'

log = logging.getLogger(__name__)

PLATFORM = platform.system().lower()

_UNC = u'\\\\?\\' if PLATFORM == 'windows' else ''

PLOT_EXECUTORS = ExecutorFactory({
    'sync': _sync_executor,
    'async': _async_executor
})
pkg_dir = osp.abspath(osp.dirname(__file__))
atexit_register(PLOT_EXECUTORS.shutdown_executor, wait=False)


def uncpath(p):
    return _UNC + osp.abspath(p)


def _encode_file_name(s, short_name=None):
    """
    Take a string and return a valid filename constructed from the string.

    Uses a whitelist approach: any characters not present in valid_chars are
    removed. Also spaces are replaced with underscores.
    """
    if short_name:
        s = hashlib.md5(s.encode()).hexdigest()[:short_name]
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    filename = ''.join(c for c in s if c in valid_chars)
    filename = filename.replace(' ', '_')  # I don't like spaces in filenames.
    return filename


def _upt_styles(styles, base=None):
    d, base = {}, copy.deepcopy(base or {})
    res = {}
    for i in ('info', 'warning', 'error'):
        combine_nested_dicts(base.get(i, {}), styles.get(i, {}), base=d)
        res[i] = copy.deepcopy(d)
    return res


def autoplot_function(kwargs):
    keys = sorted(kwargs, key=lambda x: (x is not PLOT, x))
    kw = combine_dicts(*selector(keys, kwargs, output_type='list'))
    return {'dsp': kw.pop('obj'), 'kw': kw}


def autoplot_callback(res):
    res['plot'] = res['dsp'].plot(**res['kw'])


def jinja2_format(source, context=None, **kw):
    return Environment(**kw, autoescape=True).from_string(source).render(
        context or {}
    )


def valid_filename(item, filenames, ext=None):
    if isinstance(item, str):
        _filename = item
    else:
        _filename = item._filename
        if ext is None:
            ext = item.ext
    _ = '%s' + ('.{}'.format(ext) if ext != '' else '')

    filename, c = _ % _filename, counter()
    while filename in filenames:
        filename = _ % '{}-{}'.format(_filename, c())
    return filename


def update_filenames(node, filenames):
    if node is not None:
        filename = valid_filename(node, filenames)
        yield (node, None), (filename,)
        filenames.add(filename)
        base = osp.splitext(filename)[0]
        for _, file in node.extra_files:
            filename, ext = osp.splitext(file)
            filename = valid_filename(filename, filenames, ext=ext[1:])
            yield (node, file), (osp.join(base, filename),)
            filenames.add(osp.split(filename)[0].split('.')[0])


def site_view(
        context, rendered, rules, root, filepath=None, viz=False,
        executor='async'):
    from flask import current_app
    app = current_app
    static_folder = app.static_folder
    filepath = filepath or root
    render_ctx = {
        'alive_url': 'alive' in app.view_functions and app.url_for('alive')
    }
    render_ctx.update(app.jinja_env.globals)
    if not osp.isfile(osp.join(static_folder, filepath)):
        if filepath not in rules:
            from flask import abort
            return abort(404)

        for v in cached_view(
                rules[filepath], static_folder, context, rendered, viz,
                executor, **render_ctx).values():
            v.result()
    return app.send_static_file(filepath)


def render_output(out, pformat):
    out = parent_func(out)
    if inspect.isfunction(out):
        # noinspection PyBroadException
        try:
            out = inspect.getsource(out)
        except Exception:
            pass

    if isinstance(out, (datetime.datetime, datetime.timedelta)):
        out = str(out)

    if isinstance(out, str):
        return out

    return pformat(out)


class SiteNode:
    counter = counter()
    ext = 'html'
    pprint = pprint.PrettyPrinter(compact=True, width=200)

    def __init__(
            self, folder, node_id, item, obj, dsp_node_id, short_name=None):
        self.folder = folder
        self.node_id = node_id
        self.item = item
        self.obj = obj
        self.id = str(self.counter())
        self.dsp_node_id = dsp_node_id
        self.extra_files = []
        self.short_name = short_name

    @property
    def name(self):
        try:
            return parent_func(self.item).__name__
        except AttributeError:
            return self.node_id

    @property
    def view_id(self):
        return id(self.item), self.__class__.__name__

    @property
    def title(self):
        return self.name

    @property
    def _filename(self):
        return _encode_file_name(self.title, self.short_name)

    @property
    def filename(self):
        return '.'.join((self._filename, self.ext))

    def __repr__(self):
        return self.title

    def render(self, *args, **kwargs):
        code = render_output(self.item, self.pprint.pformat)
        formatter = HtmlFormatter(noclasses=True)
        formatter.style.background_color = 'transparent'
        return _format_output(
            highlight(code, Python3Lexer(), formatter), **kwargs
        )

    def view(self, filepath, *args, **kwargs):
        fpath = uncpath(filepath)
        os.makedirs(osp.dirname(fpath), exist_ok=True)
        with open(fpath, 'w') as f:
            f.write(self.render(*args, **kwargs))
        rend = {(self.view_id, None): fpath}
        directory = osp.splitext(filepath)[0]
        for src, fn in self.extra_files:
            dst = uncpath(osp.join(directory, fn))
            os.makedirs(osp.dirname(dst), exist_ok=True)
            shutil.copy(src(**kwargs) if callable(src) else src, dst)
            rend[(self.view_id, fn)] = dst
        return rend


@functools.lru_cache(128)
def get_match_func(expr):
    return regex.compile(expr).match


@functools.lru_cache(None)
def parse_funcs(expr, funcs):
    match = get_match_func(expr)
    return [
        (f, None, None) if f == '-' or f == '?' else match(f).groups()
        for f in funcs
    ]


class FolderNode:
    counter = counter()

    node_styles = _upt_styles({
        'info': {
            START: {
                'shape': 'egg', 'label': 'start', 'fillcolor': '#39bf7f',
                'color': '#39bf7f', 'ordering': 'out'
            },
            SELF: {
                'shape': 'egg', 'label': 'self', 'fillcolor': '#C1A4FE',
                'color': '#C1A4FE'
            },
            PLOT: {
                'shape': 'egg', 'label': 'plot', 'fillcolor': '#fcf3dd',
                'color': '#fcf3dd'
            },
            END: {
                'shape': 'egg', 'label': 'end', 'fontcolor': '#FFFFFF',
                'fillcolor': '#084368', 'color': '#084368', 'ordering': 'in'
            },
            EMPTY: {
                'shape': 'egg', 'label': 'empty', 'fillcolor': '#FFFFFF'
            },
            SINK: {
                'shape': 'egg', 'label': 'sink', 'fontcolor': '#FFFFFF',
                'fillcolor': '#303030', 'color': '#303030'
            },
            NONE: {
                'data': {
                    'shape': 'box', 'style': 'rounded,filled',
                    'fillcolor': '#73c4fa', 'color': '#73c4fa'
                },
                'function': {
                    'shape': 'box', 'fillcolor': '#eed867', 'color': '#eed867',
                    'ordering': 'in'
                },
                'function-dispatcher': {
                    'shape': 'note', 'fillcolor': '#eed867', 'color': '#eed867',
                    'ordering': 'in'
                },
                'run_model': {
                    'shape': 'note', 'fillcolor': '#eed867', 'color': '#eed867',
                    'ordering': 'in'
                },
                'subdispatch': {
                    'shape': 'note', 'style': 'filled', 'fillcolor': '#ffc490',
                    'color': '#ffc490'
                },
                'mapdispatch': {
                    'shape': 'note', 'style': 'filled', 'fillcolor': '#f4bd6a',
                    'color': '#f4bd6a', 'ordering': 'in'
                },
                'subdispatchfunction': {
                    'shape': 'note', 'style': 'filled', 'fillcolor': '#f9d951',
                    'color': '#f9d951', 'ordering': 'in'
                },
                'subdispatchpipe': {
                    'shape': 'note', 'style': 'filled', 'fillcolor': '#f1cd5d',
                    'color': '#f1cd5d', 'ordering': 'in'
                },
                'dispatchpipe': {
                    'shape': 'note', 'style': 'filled', 'fillcolor': '#e8c268',
                    'color': '#e8c268', 'ordering': 'in'
                },
                'dispatcher': {
                    'shape': 'note', 'style': 'filled', 'fillcolor': '#c6c6c6',
                    'color': '#c6c6c6'
                },
                'edge': {None: None}
            }
        },
        'warning': {
            NONE: dict.fromkeys((
                'data', 'function', 'subdispatch', 'subdispatchfunction',
                'subdispatchpipe', 'dispatchpipe', 'mapdispatch', 'dispatcher',
                'run_model', 'function-dispatcher'
            ), {'fillcolor': '#fea22b', 'color': '#C9340A', 'penwidth': 2})
        },
        'error': {
            NONE: dict.fromkeys((
                'data', 'function', 'subdispatch', 'subdispatchfunction',
                'subdispatchpipe', 'dispatchpipe', 'mapdispatch', 'dispatcher',
                'run_model', 'function-dispatcher'
            ), {'fillcolor': '#FF3536', 'color': '#5E1F00', 'penwidth': 2})
        }
    })

    node_data = (
        '-', '.tooltip', '!default_values', 'wait_inputs', 'await_result',
        '+function|solution', 'weight', 'remote_links',
        '+filters|solution_filters', 'distance', '!error', '*output'
    )

    node_function = (
        '-', '.tooltip', 'await_domain', 'await_result',
        '+input_domain|solution_domain', 'weight', '+filters|solution_filters',
        'missing_inputs_outputs', 'distance', 'started', 'duration', '!error',
        '*function|solution'
    )

    edge_data = '?', '+wildcard', 'inp_id', 'out_id', 'weight'

    node_map = {
        '-': (),  # Add title.
        '?': (),  # Optional title.
        '': ('dot', 'table'),  # Item in the table.
        '+': ('dot', 'table'),  # Link.
        '!': ('dot', 'table'),  # If str is big add a link, otherwise table.
        '.': ('dot',),  # Dot attr.
        '*': ('link',)  # Title link.
    }
    re_node = r"^([.*+!]?)([\w ]+)(?>\|([\w ]+))?$"
    max_lines = 5
    max_width = 200
    pprint = pprint.PrettyPrinter(compact=True, width=200)

    def __init__(self, folder, node_id, attr, **options):
        self.folder = folder
        self.node_id = node_id
        self.attr = attr
        self.id = str(self.counter())
        self._links = {}
        for k, v in options.items():
            setattr(self, k, v)

    @property
    def title(self):
        return self.attr.get('title', self.node_id)

    @property
    def type(self):
        return self.attr.get('type', 'data')

    def __repr__(self):
        return self.title

    def yield_attr(self, name):
        try:
            yield name, self.attr[name]
        except KeyError:
            pass

    def render_size(self, out):
        lines = render_output(out, self.pprint.pformat).splitlines(True)
        n, w = self.max_lines, self.max_width
        return len(lines) <= n and not any(len(v) > w for v in lines)

    def items(self):
        check = self.render_size
        for k, func in self.render_funcs():
            if k and k in '*+':
                yield from func()
            elif k == '!':
                yield from ((i, j) for i, j in func() if not check(j))

    def _tooltip(self):
        try:
            from ..des import search_node_description
            tooltip = search_node_description(
                self.node_id, self.attr, self.folder.dsp
            )[0]
        except (AttributeError, KeyError):
            tooltip = None
        yield 'tooltip', '"%s"' % (tooltip or self.title).replace('"', "'")

    def _wait_inputs(self):
        attr = self.attr
        try:
            if attr['type'] == 'data' and attr['wait_inputs']:
                yield 'wait_inputs', attr['wait_inputs']
        except KeyError:
            pass

    def _default_values(self):
        try:
            dfl = self.folder.dsp.default_values.get(self.node_id, {})
            res = map_dict({'value': 'default'}, dfl)

            if not res.get('initial_dist', 1):
                res.pop('initial_dist')
        except AttributeError:
            res = {}
        yield from sorted(res.items())

    def _filters(self):
        try:
            for i, f in enumerate(self.attr['filters']):
                yield 'filter %d' % i, f
        except (AttributeError, KeyError):
            pass

    def _solution_filters(self):
        try:
            it = self.attr['solution_filters']
            yield 'input_filter 0', it[0]
            for i, f in enumerate(it[1:]):
                yield 'output_filter %d' % i, f
        except (AttributeError, KeyError, IndexError):
            pass

    def _remote_links(self):
        if not self.folder.parent:
            return
        dsp_id, dsp = self.folder.name, self.folder.parent.dsp
        node = dsp.nodes[dsp_id]
        if node['type'] == 'dispatcher':
            i, nid = 0, self.node_id
            for k, tag in (('inputs', 'parent'), ('outputs', 'child')):
                d = node.get(k, {})
                if tag == 'parent':
                    n = tuple(k for k, v in d.items() if nid in stlp(v))
                else:
                    n = stlp(d.get(nid, ()))

                if not n:
                    continue
                elif len(n) == 1:
                    n = n[0]

                n = 'parent_ref("{}", attr)'.format(n)
                yield 'remote %s %d' % (tag, i), '{{%s}}' % n
                i += 1

    def _output(self):
        if self.node_id not in (START, SINK, SELF, END):
            try:
                out = self.folder.item[self.node_id]
                yield 'output', out
            except (KeyError, TypeError):
                pass  # Output not in solution or item is not a solution.

    def _started(self):
        try:
            if isinstance(self.attr['started'], str):
                yield 'started', self.attr['started']
            else:
                started = datetime.datetime.fromtimestamp(self.attr['started'])
                yield 'started', started.isoformat()
        except KeyError:
            pass

    def _duration(self):
        k = 'duration'
        try:
            if isinstance(self.attr[k], str):
                yield k, self.attr[k]
            else:
                yield k, datetime.timedelta(seconds=self.attr[k])
        except KeyError:
            pass

    def _distance(self):
        try:
            yield 'distance', self.folder.item.dist[self.node_id]
        except (AttributeError, KeyError):
            yield from self.yield_attr('distance')

    def _weight(self):
        try:
            yield 'weight', self.attr['weight']
        except (AttributeError, KeyError):
            pass

    def _missing_inputs_outputs(self):
        attr, res = self.attr, {}
        try:
            if attr['wait_inputs']:
                graph = self.folder.graph
                pred, succ = graph.pred[self.node_id], graph.succ[self.node_id]
                for i, j in (('inputs', pred), ('outputs', succ)):
                    v = tuple(k for k in attr[i] if k not in j)
                    if v:
                        yield 'M_%s' % i, v
        except (AttributeError, KeyError):
            pass

    def _wildcard(self):
        if 'wildcard' in self.attr:
            if self.folder.workflow and 'value' in self.attr:
                yield 'wildcard', self.attr['value']

    def style(self):
        attr = self.attr
        if 'style' in attr:
            return attr['style'].copy()
        elif 'error' in attr:
            nstyle = 'error'
        elif list(self._missing_inputs_outputs()):
            nstyle = 'warning'
        else:
            nstyle = 'info'

        node_styles = self.node_styles.get(nstyle, self.node_styles['info'])
        if self.node_id in node_styles:
            node_style = node_styles[self.node_id].copy()
            node_style.pop(None, None)
            return node_style
        else:
            if self.type in ('dispatcher', 'function'):
                ntype = 'function',
                try:
                    ntype = (_get_type(attr['function'], self.type),) + ntype
                except (KeyError, AttributeError):
                    pass
            elif self.type == 'edge':
                ntype = 'edge',
            else:
                ntype = 'data',
            for style in ntype:
                try:
                    node_style = node_styles[NONE][style].copy()
                    node_style.pop(None, None)
                    return node_style
                except KeyError:
                    pass

    def render_funcs(self):
        if self.type in ('dispatcher', 'function'):
            funcs = self.node_function
        elif self.type == 'edge':
            funcs = self.edge_data
        else:
            funcs = self.node_data
        r, s = {}, '_%s'
        workflow = self.folder.workflow
        for k, v, v1 in parse_funcs(self.re_node, funcs):
            if k == '-' or k == '?':
                yield k, lambda *args: self.title
            else:
                if workflow and v1:
                    try:
                        yield k, getattr(self, s % v1)
                        continue
                    except AttributeError:
                        if v1 in self.attr:
                            yield k, functools.partial(self.yield_attr, v1)
                            continue
                try:
                    yield k, getattr(self, s % v)
                except AttributeError:
                    yield k, functools.partial(self.yield_attr, v)

    def parent_ref(self, context, node_id, attr=None):
        attr, text = attr or {}, '(%s)' % node_id
        try:
            dirname = osp.dirname(context[(self.folder, None)])
            node, rule = next((n, f) for (n, e), f in context.items()
                              if e is None and dirname == osp.splitext(f)[0])
            attr, href = attr.copy(), osp.relpath(rule, dirname)
            node_id = next(
                '?id=%d' % n.attr['index'][-1]
                for n in node.nodes
                if n.node_id == node_id
            )
            attr['href'] = html.escape(urlparse.unquote('./%s%s' % (
                href.replace('\\', '/'), node_id or ''
            )))
        except StopIteration:
            pass

        return '_Td(**{}).add("{}")'.format(attr, text)

    def href(self, context, link_id):
        res = {}
        if link_id in self._links:
            node = self._links[link_id]
            res['text'] = node.title
            try:
                dirname = osp.dirname(context[(self.folder, None)])
                href = osp.relpath(context[(node, None)], dirname)
                res['href'] = urlparse.unquote('./%s' % href.replace('\\', '/'))
            except KeyError:
                pass
        return res

    def dot(self, context=None):
        if context is None:
            context = {}
        dot = self.style()
        if 'label' in dot:
            dot.update(self.attr.get('graphviz', {}))
            return dot
        from .nodes import _Tr, _Td
        key = {'ALIGN': "RIGHT", 'BORDER': 1}
        val = {'ALIGN': "LEFT", 'BORDER': 1}
        rows, funcs, cnt = [], list(self.render_funcs()), {'attr': val}
        cnt['parent_ref'] = functools.partial(self.parent_ref, context)
        href, pformat, links = self.href, self.pprint.pformat, self._links
        for k, func in funcs:
            if k == '.':
                dot.update(func())
            elif not (k == '*' or k == '-' or k == '?'):

                for i, j in func():
                    tr = _Tr().add(i, **key)
                    if i in links and (k == '!' or k == '+'):
                        v = combine_dicts(val, {'text': j}, href(context, i))
                        tr.add(**v)
                    else:
                        j = render_output(j, pformat)
                        s = html.unescape(jinja2_format(j, cnt))
                        if s.startswith('_Td('):
                            tr += eval(s)
                        else:  # It is not a valid jinja2 format.
                            tr.add(j, **val)
                    rows.append(tr)

        def get_link():
            for _k, f in funcs:
                if _k == '*':
                    for _link_id in f():
                        return _link_id[0]

        if any(k[0] == '-' or (rows and k[0] == '?') for k in funcs):
            link_id = get_link()
            kw = combine_dicts(
                self.href(context, link_id),
                {'COLSPAN': 2, 'BORDER': 0, 'text': self.title}
            )
            rows = [_Tr().add(**kw)] + rows

        if rows:
            k = 'xlabel' if self.type == 'edge' else 'label'
            k = self.attr.get('label_type', k)
            from .nodes import _Table
            dot[k] = '<%s>' % _Table(
                BORDER=0, CELLSPACING=0, COLOR=dot.get('fontcolor', 'black')
            ).adds(rows)
        dot.update(self.attr.get('graphviz', {}))
        return {k: str(v) for k, v in dot.items()}


def _format_output(obj, **kwargs):
    fpath = osp.join(pkg_dir, 'templates', 'render.html')
    with open(fpath) as template:
        return jinja2_format(
            template.read(), combine_dicts(kwargs, base={'obj': obj}),
            loader=PackageLoader(__name__)
        )


def _format_kw_digraph(*dicts, base=None):
    kw = combine_nested_dicts(*dicts, base=base)
    body = kw.pop('body', None)
    if 'raw_body' in kw:
        get_nested_dicts(kw, 'body', default=list).extend(kw.pop('raw_body'))
    if body:
        get_nested_dicts(kw, 'body', default=list).extend(
            '\t%s = %s\n' % (k, v) for k, v in sorted(body.items())
        )
    return kw


class SiteFolder:
    counter = SiteNode.counter
    digraph = {
        'node_attr': {'style': 'filled'},
        'graph_attr': {
            'bgcolor': 'transparent', 'nslimit': '1', 'nslimit1': '1'
        },
        'edge_attr': {},
        'body': {'splines': 'ortho', 'style': 'filled'},
        'format': 'svg'
    }
    folder_node = FolderNode
    ext = 'html'

    def __init__(self, item, dsp, graph, obj, name='', workflow=False,
                 digraph=None, parent=None, short_name=None, **options):
        self.item, self.dsp, self.graph, self.obj = item, dsp, graph, obj
        self._name = name
        self.workflow = workflow
        self.parent = parent
        self.id = str(self.counter())
        self.options = options
        nodes = collections.OrderedDict(self._nodes)
        self.nodes = list(nodes.values())
        self.edges = [e for k, e in self._edges(nodes)]
        self.sitemap = None
        self.extra_files = []
        self.short_name = short_name
        if digraph is not None:
            self.digraph = combine_dicts(self.__class__.digraph, digraph)

    @property
    def view_id(self):
        return id(self.item), self.__class__.__name__

    @property
    def title(self):
        return self.name or ''

    @property
    def _filename(self):
        return _encode_file_name(self.title, self.short_name)

    @property
    def filename(self):
        return '.'.join((self._filename, self.ext))

    def __repr__(self):
        return self.title

    @property
    def inputs(self):
        try:
            from ..sol import Solution
            if isinstance(self.item, Solution):
                return self.item.parent.inputs or ()
            return self.item.inputs or ()
        except AttributeError:
            return ()

    @property
    def outputs(self):
        item = self.item
        from ..sol import Solution
        try:
            if isinstance(item, Solution):
                item = item.parent
            if isinstance(item, SubDispatch) and item.output_type != 'all':
                return item.outputs or ()
        except AttributeError:
            pass
        return ()

    @property
    def name(self):
        if not self._name:
            dsp = self.dsp
            name = dsp.name or '%s %d' % (type(dsp).__name__, id(dsp))
        else:
            name = self._name
        return name

    @property
    def label_name(self):
        return 'workflow' if self.workflow else 'dmap'

    @property
    def _nodes(self):
        try:
            errors = self.item._errors
        except AttributeError:
            errors = {}
        nodes, graph = self.dsp.nodes, self.graph
        gnode, succ, pred = graph.nodes, graph.succ, graph.pred
        it = {
            i: v for i, v in gnode.items()
            if i in nodes and (i is not SINK or succ[SINK] or pred[SINK])
        }
        if not nodes and not (graph.edges or self.inputs or self.outputs):
            it[EMPTY] = {'index': (EMPTY,)}

        if START in gnode or any(i in it for i in self.inputs):
            it[START] = {'index': (START,)}

        if any(o in it for o in self.outputs) and END not in gnode:
            it[END] = {'index': (END,)}
        it = sorted(
            it.items(),
            key=lambda x: (
                x[0] is END, nodes.get(x[0], {}).get('index', ()), x[0])
        )
        for k, a in it:
            attr = combine_dicts(nodes.get(k, {}), a)
            if k in errors:
                attr['error'] = errors[k]

            yield k, self.folder_node(self, k, attr, **self.options)

    def _edges(self, nodes):
        edges = {(u, v): a for (u, v), a in self.graph.edges.items() if u != v}
        from ..sol import Solution
        from ..dsp import SubDispatchFunction
        wildcards = ()
        if isinstance(self.item, Solution):
            wildcards = self.item._wildcards
        elif isinstance(self.item, SubDispatchFunction):
            wildcards = self.item._sol._wildcards

        for i, v in enumerate(self.inputs):
            if v != START and v in nodes:
                n = (START, v)
                edges[n] = combine_dicts(edges.get(n, {}), {'inp_id': i})

        for i, u in enumerate(self.outputs):
            if u != END and u in nodes:
                n = (u, END)
                edges[n] = combine_dicts(edges.get(n, {}), {'out_id': i})

        for w in [v for u, v in edges if u is START and v in wildcards]:
            a = combine_dicts(edges.pop((START, w)), {
                'label_type': 'label', 'wildcard': w, 'title': w
            })
            a1 = selector(('value',), a, allow_miss=True)
            for u, v in list(edges):
                if u == w and v != END:
                    edges[(START, v)] = combine_dicts(a, edges.pop((u, v)), a1)

        d_nodes = self.dsp.nodes
        for (u, v), a in edges.items():
            base = {'type': 'edge', 'dot_ids': (nodes[u].id, nodes[v].id)}
            a = combine_dicts(a, base=base)
            if v in d_nodes and d_nodes[v]['type'] == 'dispatcher':
                if not a.get('weight', 1):
                    a.pop('weight')
            elif a.get('weight') == 1:
                a.pop('weight')

            yield (u, v), self.folder_node(self, '{} --> {}'.format(u, v), a)

    def dot(self, context=None):
        context = context or {}

        kw = _format_kw_digraph(self.digraph, base={
            'name': self.label_name, 'body': {'label': '<%s>' % self.label_name}
        })
        from .nodes import _DspPlot
        dot = _DspPlot(self.sitemap, **kw)
        id_map, clr = {}, {}
        for node in self.nodes:
            i = id_map[node.node_id] = node.id
            dot.node(i, id=str(node.attr['index'][-1]), **node.dot(context))
            clusters = node.attr.get('clusters', ())
            if isinstance(clusters, (str, dict)):
                clusters = clusters,
            for c in clusters:
                if isinstance(c, dict):
                    j = c['body']['label']
                    combine_nested_dicts(c, base=get_nested_dicts(clr, j, 'kw'))
                    if j and j[0] == '<' and j[-1] == '>':
                        j = j[1:-1]
                else:
                    j = c
                get_nested_dicts(clr, j, 'nodes', default=list).append(i)

        for edge in self.edges:
            dot.edge(*edge.attr['dot_ids'], **edge.dot(context))

        for i, (cluster, d) in enumerate(clr.items()):
            kw = _format_kw_digraph(d.get('kw', {}), base={
                'name': 'cluster_%d' % i, 'body': {'label': '"%s"' % cluster}
            })
            with dot.subgraph(**kw) as g:
                for node in d['nodes']:
                    g.node(node)
        return dot

    # noinspection PyUnusedLocal
    def view(self, filepath, context=None, viz=False, **kwargs):
        dot = self.dot(context=context)
        dot.format = self.digraph['format']
        filepath = uncpath(filepath)
        if osp.isfile(filepath):
            os.remove(filepath)
        else:
            os.makedirs(osp.dirname(filepath), exist_ok=True)
        if viz and dot.format == 'svg':
            out = '<viz engine="%s" digraph="%s"/>'
            out %= dot.engine, html.escape(dot.source)
        else:
            try:
                # noinspection PyArgumentList
                fpath = dot.render(directory=tempfile.mkdtemp(), cleanup=True)
                with open(fpath) as src:
                    out = src.read()
                os.remove(fpath)
            except Exception as ex:
                raise ValueError('dot could not render %s (%s) due to:\n %r' % (
                    filepath, dot.filepath, ex
                ))
        with open(filepath, 'w') as dst:
            if viz:
                viz = len(context[(self, None)].split(osp.sep)) - 1
                viz = '/'.join(('..',) * viz + ('viz.js',))
            dst.write(_format_output(out, viz=viz, **kwargs))
        return {(self.view_id, None): filepath}


def _get_type(obj, node_type='function'):
    from ..sol import Solution
    obj = parent_func(isinstance(obj, Solution) and obj.parent or obj)
    if not isinstance(obj, type):
        obj = type(obj)
    typ = obj.__name__.lower()
    if typ == 'dispatcher' and node_type == 'function':
        return 'function-dispatcher'
    return typ


sort_tree_map = {v: k for k, v in enumerate((
    'data', 'function', 'run_model', 'dispatchpipe', 'subdispatchpipe',
    'subdispatchfunction', 'mapdispatch', 'subdispatch', 'dispatcher'
), 1)}


def _folder2tree(folder, smap, context, type):
    extra, extra_dsp = {}, {item.name: (item, v) for item, v in smap.items()}
    for item in smap.nodes:
        get_nested_dicts(extra, item.dsp_node_id, default=list).append(item)
    url = context[(folder, None)]
    nodes = [{'text': '-%s' % type, 'url': url, 'type': type}]
    url = '{}?id=%d'.format(url)
    for node_id, attr in folder.dsp.nodes.items():
        if node_id not in folder.graph.nodes:
            continue
        type = attr['type']
        if type == 'function' and attr.get('function'):
            type = _get_type(attr['function'])
        n = {
            'text': html.escape(node_id), 'url': url % attr['index'][-1],
            'type': type
        }
        nodes.append(n)
        if node_id in extra_dsp:
            n['nodes'] = _folder2tree(*extra_dsp[node_id], context, type)
        else:
            i = len(node_id)
            n['nodes'] = [{
                'text': html.escape(item.title[i:] or '-function'),
                'url': context[(item, None)],
                'type': _get_type(item.item)
            } for item in extra.get(node_id, [])]
        if not n['nodes']:
            n.pop('nodes')
    return nodes[:1] + list(sorted(
        nodes[1:], key=lambda x: sort_tree_map.get(x['type'], 0), reverse=True
    ))


def _pipe2icicle(pipe):
    for k, v in pipe.items():
        child, (i, s) = {'name': ' → '.join(stlp(k))}, v['task'][2]
        value, t = 0, s.workflow.nodes.get(i, {}).get('duration')
        if 'sub_pipe' in v:
            child['children'] = children = list(_pipe2icicle(v['sub_pipe']))
            dt = sum((v['duration'] for v in children), 0)
            if t is None:
                t = dt
            else:
                value = t - dt
        else:
            value = t or 0
        child['duration'] = t or 0
        child['value'] = value
        yield child


def _sitemap2icicle(sitemap):
    cdn = []
    for folder in sitemap:
        try:
            pipe = folder.item.pipe
            assert isinstance(pipe, dict)
        except (AttributeError, AssertionError):
            continue

        c = list(_pipe2icicle(pipe))
        cdn.append({
            'name': folder.name, 'children': c, 'value': 0,
            'duration': sum((v['duration'] for v in c), 0)
        })
    if not cdn:
        return {}
    return {
        'name': 'main', 'duration': sum((v['duration'] for v in cdn), 0),
        'children': cdn
    }


def _sitemap2tree(sitemap, context):
    tree = []
    for folder, smap in sitemap.items():
        type = _get_type(folder.item)
        tree.append({
            'text': html.escape(folder.name),
            'url': context[(folder, None)],
            'type': type,
            'nodes': _folder2tree(folder, smap, context, type)[1:]
        })
    return tree


def _add_explanation(dsp, node_id, description, **kw):
    dsp.dmap.add_edge(node_id, dsp.add_data(graphviz={
        'label': description, 'shape': 'plaintext', 'style': '', 'fillcolor': ''
    }, **kw), graphviz={'style': 'dashed'})
    return node_id


class NoView:
    pass


class SiteViz(SiteNode, NoView):
    ext = 'js'

    def __init__(self, sitemap, node_id='viz'):
        super(SiteViz, self).__init__(None, node_id, self, None, object())
        self.sitemap = sitemap

    def render(self, context, *args, **kwargs):
        with open(osp.join(pkg_dir, 'viz', 'viz.js')) as f:
            return f.read()


class SiteIndex(SiteNode):
    ext = 'html'

    def __init__(self, sitemap, node_id='index'):
        super(SiteIndex, self).__init__(None, node_id, self, None, object())
        self.sitemap = sitemap
        dfl_folder = osp.join(pkg_dir, 'index')
        for default_file in glob.glob(osp.join(dfl_folder, '**/*')):
            if osp.isfile(default_file):
                self.extra_files.append(
                    (default_file, osp.relpath(default_file, dfl_folder))
                )
        self.extra_files.append((self.legend, 'html/legend.html'))

    # noinspection PyUnusedLocal
    @staticmethod
    def legend(viz=False, executor='async', **kwargs):
        import schedula as sh
        dsp = sh.Dispatcher(name='legend')
        _add_explanation(dsp, dsp.add_data(
            clusters='Data',
            data_id='<data_id>(Data)',
            default_value='Default value.',
            initial_dist='Initial distance from `START` node.',
            wait_inputs='Wait all data estimations? `<bool>`',
            function='Process data function `data = f({"<node name>": data})`.',
            callback='Callback function `f(data)`.',
            await_result='Wait async for function result? `<bool>`',
            distance='Distance from `START` node.',
            started='Execution started time.',
            duration='Time elapsed to execute all functions '
                     'that belong to the node.',
            **{'remote parent x': 'Link to a node of the parent dispatcher.',
               'remote child x': 'Link to a node ot a child dispatcher.',
               'filter x': 'Filter function x used in the loop '
                           '`for f in filters: data = f(data)`.',
               'input_filter 0': 'Result of the function.',
               'output_filter x': 'Result of filters[x](`input_filter x` | '
                                  '`output_filter x-1`).'}
        ), clusters='Data', description=(
            'Data node. The current one is a\n'
            'sample showing the main attributes.'
        ))
        _add_explanation(
            dsp, dsp.add_data(sh.EMPTY, clusters='Special Nodes'),
            'Empty dispatcher/workflow.', clusters='Special Nodes'
        )
        _add_explanation(
            dsp, dsp.add_data(sh.START, clusters='Special Nodes'),
            'Starting node. It identifies'
            '\nthe initial inputs.', clusters='Special Nodes'
        )
        _add_explanation(
            dsp, dsp.add_data(sh.END, clusters='Special Nodes'),
            'Ending node of SubDispatcherFunction.\n'
            'It collects the function\'s outputs.', clusters='Special Nodes'
        )
        _add_explanation(
            dsp, dsp.add_data(sh.SINK, clusters='Special Nodes'),
            'Sink node. It collects \n'
            'all unused outputs.', clusters='Special Nodes'
        )
        _add_explanation(
            dsp, dsp.add_data(sh.SELF, clusters='Special Nodes'),
            'Self node of the plotted dispatcher.\n'
            'It represents the dispatcher as data node.',
            clusters='Special Nodes'
        )
        _add_explanation(
            dsp, dsp.add_data(sh.PLOT, clusters='Special Nodes'),
            'Plot node. When invoked, it\n'
            'plots the dispatcher solution.', clusters='Special Nodes'
        )

        fun_kw = {'inputs': [], 'outputs': [], 'clusters': 'Functions'}

        class subdispatch:
            pass

        class run_model:
            pass

        class mapdispatch:
            pass

        class subdispatchfunction:
            pass

        class subdispatchpipe:
            pass

        class dispatchpipe:
            pass

        _add_explanation(dsp, dsp.add_function(
            function_id='<function_id>(Function)',
            input_domain='Domain function `f(*inputs)`.',
            solution_domain='Domain function result.',
            weight='Distance weight coeff.',
            await_domain='Wait async for domain result? `<bool>`',
            await_result='Wait async for function result? `<bool>`',
            distance='Distance from `START` node.',
            started='Execution started time.',
            duration='Time elapsed to execute the function.',
            **fun_kw,
            **{'filter x': 'Filter function x used in the loop '
                           '`for f in filters: output = f(output)`.',
               'input_filter 0': 'Result of the function.',
               'output_filter x': 'Result of filters[x](`input_filter x` | '
                                  '`output_filter x-1`).'}
        ), clusters='Functions', description=(
            'Function node. The current one is a\n'
            'sample showing the main attributes.'
        ))

        _add_explanation(dsp, dsp.add_function(
            function_id='<function_id>(run_model)',
            function=run_model(), **fun_kw
        ), clusters='Functions', description=(
            'run_model node. It calls the first argument - i.e. the model.\n '
            'If the model produces a the workflow, this is added to the\n'
            'parent workflow.\n'
            'First argument is a callable.'
        ))
        _add_explanation(dsp, dsp.add_function(
            function_id='<function_id>(SubDispatch)',
            function=subdispatch(), **fun_kw
        ), clusters='Functions', description=(
            'SubDispatch node. It wraps\n'
            'a given Dispatcher into a function.\n'
            'Inputs are dictionaries {<node_id>: <value>}.'
        ))

        _add_explanation(dsp, dsp.add_function(
            function_id='<function_id>(MapDispatch)',
            function=mapdispatch(), **fun_kw
        ), clusters='Functions', description=(
            'MapDispatch node. It wraps and executes iteratively\n'
            'a given Dispatcher into a function.\n'
            'Hence, it behaves like a `map` function.'
        ))
        _add_explanation(dsp, dsp.add_function(
            function_id='<function_id>(SubDispatchFunction)',
            function=subdispatchfunction(), **fun_kw
        ), clusters='Functions', description=(
            'SubDispatchFunction node. It wraps and shrink\n'
            'a given Dispatcher into a function.\n'
            'Hence, it behaves like a function.'
        ))
        _add_explanation(dsp, dsp.add_function(
            function_id='<function_id>(SubDispatchPipe)',
            function=subdispatchpipe(), **fun_kw
        ), clusters='Functions', description=(
            'SubDispatchPipe node. It wraps and compiles\n'
            'a given Dispatcher into a function.\n'
            'Hence, it behaves like a function.'
        ))
        _add_explanation(dsp, dsp.add_function(
            function_id='<function_id>(DispatchPipe)',
            function=dispatchpipe(), **fun_kw
        ), clusters='Functions', description=(
            'DispatchPipe node. It behaves like a\n'
            'SubDispatchPipe node, but it overwrites\n'
            'its solution.'
        ))

        _add_explanation(dsp, dsp.add_function(
            function_id='<node_id>(Error)',
            clusters='Warnings and Errors', inputs=[], outputs=[],
            error='Error message.'
        ), clusters='Warnings and Errors', description=(
            'Node that raised an error during its execution.'
        ))
        d = dsp.nodes[_add_explanation(dsp, dsp.add_function(
            function_id='<node_id>(Warning)',
            clusters='Warnings and Errors', inputs=[], outputs=[]
        ), clusters='Warnings and Errors', description=(
            'Node that did not return all inputs/outputs.'
        ))]
        d['outputs'].append('missing outputs')
        d['inputs'].append('missing inputs')

        _add_explanation(dsp, dsp.add_dispatcher(
            clusters='Sub-dispatcher',
            dsp={},
            inputs={},
            outputs={},
            dsp_id='<dispatcher_id>(Dispatcher)',
            input_domain='Domain function `f(**inputs)`.',
            solution_domain='Domain function result.',
            weight='Distance weight coeff.',
            await_domain='Wait async for domain result? `<bool>`',
            distance='Distance from `START` node.',
        ), clusters='Sub-dispatcher', description=(
            'Sub-dispatcher node. It connects a given\n'
            'dispatcher to the current one. The current\n'
            'node one is a sample showing the main attributes.'
        ))

        dsp.add_data('<from>', clusters='Edges', graphviz={'style': 'invis'})
        dsp.add_function(
            function_id='<to>',
            clusters='Edges',
            inputs=['<from>'],
            outputs=[],
            inp_weight={'<from>': 'Edge distance.'},
            graphviz={'style': 'invis'}
        )
        dsp.dmap['<from>']['<to>'].update({
            'label_type': 'label',
            'inp_id': 'Index of input args.',
            'out_id': 'Index of output list.',
            'graphviz': {
                'xlabel': 'This is an edge sample showing the main attributes.'
            }
        })
        return dsp.plot(
            view=False,
            name='legend',
            body={'label': '""'},
            graph_attr={'rankdir': 'LR', 'id': 'graph'},
            node_data=(
                '-', 'default_values', 'wait_inputs',
                'await_result', 'distance', 'function', 'solution', 'weight',
                'remote_links', 'filter x', 'input_filter 0', 'output_filter x',
                'error', '*output', 'remote parent x', 'remote child x',
            ),
            node_function=(
                '-', 'await_domain', 'await_result', 'input_domain',
                'solution_domain', 'weight', 'filter x', 'input_filter 0',
                'output_filter x', 'error', 'missing_inputs_outputs',
                'distance', 'started', 'duration',
            )
        ).render(
            view=False, index=False, directory=tempfile.mkdtemp(), viz=viz,
            executor=executor
        )

    def render(self, context, *args, **kwargs):
        fpath = osp.join(pkg_dir, 'templates', 'index.html')

        with open(fpath) as template:
            return jinja2_format(
                template.read(),
                {'tree': _sitemap2tree(self.sitemap, context),
                 'icicle': _sitemap2icicle(self.sitemap),
                 'pid': threading.get_ident(),
                 'folder': self._filename},
                loader=PackageLoader(__name__)
            )


def run_server(app, options):
    MUTE_REQUESTS[threading.get_ident()] = getattr(app, 'mute', False)
    app.run(**options)


MUTE_REQUESTS = {}
logging.getLogger('werkzeug').addFilter(
    lambda r: not MUTE_REQUESTS.pop(r.thread, False)
)


def before_request(mute):
    import flask
    import threading
    from flask import request
    MUTE_REQUESTS[threading.get_ident()] = mute
    method = request.form.get('_method', '').upper()
    if method:
        request.environ['REQUEST_METHOD'] = method
        ctx = flask._request_ctx_stack.top
        ctx.url_adapter.default_method = method
        assert request.method == method


# noinspection HttpUrlsUsage
_repr_html = '''
<style> .sh-box {{ width: 100%; height: 500px }} </style>
<iframe id="{id}" class="sh-box" src="http://{host}:{port}/" allowfullscreen>
</iframe>
'''


class ServerThread(threading.Thread):

    def __init__(self, application, threaded=True, **kwargs):
        from werkzeug.serving import make_server
        threading.Thread.__init__(self)
        self.srv = make_server(app=application, threaded=threaded, **kwargs)

    def run(self):
        log.info('starting server')
        self.srv.serve_forever()

    def shutdown(self):
        log.info('shutdown server')
        self.srv.shutdown()


class IdleContainer(threading.Thread):
    def __init__(self, interval=1):
        super(IdleContainer, self).__init__(daemon=True)
        self.sites = set()
        self.interval = interval
        self.stop = threading.Event()

    def add(self, site):
        self.sites.add(site)
        if not self._started.is_set():
            self.start()

    def run(self):
        dt = self.interval + 2
        while not self.stop.is_set():
            time.sleep(self.interval)
            t = time.time() - dt
            for site in tuple(self.sites):
                if (site.idle_timeout + site.last_activity) < t:
                    site.shutdown()
                    self.sites.remove(site)


IDLES = IdleContainer()


class ReverseProxied:
    def __init__(self, app, script_name):
        self.app = app
        self.script_name = script_name

    def __call__(self, environ, start_response):
        environ['SCRIPT_NAME'] = self.script_name
        return self.app(environ, start_response)


class Site:
    def __init__(self, sitemap, host='localhost', port=0, delay=0.1, until=30,
                 run_options=None, idle_timeout=0, url_prefix=None, **kwargs):
        self.sitemap = sitemap
        self.kwargs = kwargs
        self.host = host
        self.port = port
        self.shutdown = lambda: False
        self.delay = delay
        self.until = until
        self.idle_timeout = idle_timeout
        self._html = os.environ.get("SCHEDULA_SITE_REPR_HTML", _repr_html)
        self.run_options = {} if run_options is None else run_options
        if url_prefix is None:
            self.proxy = os.environ.get("SCHEDULA_FORM_URL_PREFIX", '')
        else:
            self.proxy = url_prefix

    def __repr__(self):
        s = "%s(%s, " % (self.__class__.__name__, self.sitemap)
        s += "host='{}', port={}".format(self.host, self.port)
        for k, v in sorted(self.kwargs.items()):
            s += ', {}={}'.format(k, ("'%s'" % v) if isinstance(v, str) else v)
        return s + ')'

    def get_port(self, host=None, port=None, **kw):
        kw = kw.copy()
        kw['host'] = self.host = host or self.host
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.host, port or self.port))
        kw['port'] = self.port = sock.getsockname()[1]
        sock.close()
        return kw

    def format(self, string):
        # noinspection PyTypeChecker
        kw = combine_dicts(os.environ, {
            'id': id(self), 'host': self.host, 'port': self.port
        })
        return string.format(**kw)

    def _repr_html_(self):
        from IPython.display import HTML
        if not self.is_running:
            self.run()
        return HTML(self.format(self._html))._repr_html_()

    @property
    def url(self):
        # noinspection HttpUrlsUsage
        return 'http://{}:{}'.format(self.host, self.port)

    def update_last_activity(self, response=None):
        self.last_activity = time.time()
        return response

    @staticmethod
    def alive_view():
        return 'active'

    def app(self):
        app = self.sitemap.app(**self.kwargs)

        if self.idle_timeout > 0:
            app.after_app_request(self.update_last_activity)
            if 'alive' not in app.view_functions:
                app.add_url_rule('/alive', 'alive', self.alive_view)
            app.add_app_template_global(self.idle_timeout, 'alive_period')
        from flask import Blueprint, Flask
        if isinstance(app, Blueprint):
            bp = app
            app = Flask(
                import_name=bp.import_name,
                static_folder=bp.static_folder or 'static',
                static_url_path=bp.static_url_path,
                template_folder=bp.template_folder or 'templates',
                root_path=bp.root_path
            )
            app.register_blueprint(bp)
        if self.proxy:
            app.wsgi_app = ReverseProxied(
                app.wsgi_app, script_name=self.format(self.proxy)
            )
        return app

    @staticmethod
    def shutdown_site(shutdown, subsites):
        for _, site_shutdown in subsites.values():
            site_shutdown()
        subsites.clear()
        shutdown()
        return True

    def run(self, **options):
        self.shutdown()
        options = combine_dicts(self.run_options, options)
        memo = os.environ.get("WERKZEUG_RUN_MAIN")
        try:
            os.environ["WERKZEUG_RUN_MAIN"] = "true"
            kw = self.get_port(**options)
            app = self.app()
            thread = ServerThread(app, **kw)
            thread.start()
            # noinspection PyArgumentList
            self.shutdown = weakref.finalize(
                self, self.shutdown_site, thread.shutdown,
                self.sitemap and self.sitemap.subsites or {}
            )
            self.wait_server()
            if self.idle_timeout > 0:
                IDLES.add(self)
                self.update_last_activity()
            time.sleep(max(min(self.delay, 1), .1))
        finally:
            if memo is None:
                os.environ.pop("WERKZEUG_RUN_MAIN")
            else:
                os.environ["WERKZEUG_RUN_MAIN"] = memo

        return self

    @property
    def is_running(self):
        running = False
        if self.port:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:  # Tries to connect to the host.
                sock.connect((self.host, self.port))
                running = True
            except ConnectionRefusedError:
                pass
            finally:
                sock.close()
        return running

    def wait_server(self, elapsed=0):
        end = time.time() + self.until
        while not self.is_running:
            time.sleep(self.delay)
            if time.time() > end:
                msg = 'After %.3fs the server %s is down!' % (elapsed, self.url)
                raise ConnectionRefusedError(msg)


class SiteMap(collections.OrderedDict):
    site_folder = SiteFolder
    site_node = SiteNode
    site_index = SiteIndex
    site_viz = SiteViz
    short_name = None
    directory = None
    blueprint_name = None
    _view = None
    options = {
        'digraph', 'node_styles', 'node_data', 'node_function', 'edge_data',
        'max_lines', 'max_width'
    }
    include_folders_as_filenames = True

    def __init__(self):
        super(SiteMap, self).__init__()
        if self._view is None:
            from .nodes import _DspPlot
            self._view = _DspPlot(None)._view
        self._nodes = []
        self.foldername = ''
        self.filenames = {'index', 'index.html', 'viz.js'}
        self.subsites = {}

    def __setitem__(self, key, value, *args, **kwargs):
        value.foldername = valid_filename(key, self.filenames, ext='')
        self.filenames.add(value.foldername)
        # noinspection PyArgumentList
        super(SiteMap, self).__setitem__(key, value, *args, **kwargs)

    def _repr_svg_(self):
        dot = list(self)[-1].dot()
        return dot.pipe(format='svg').decode(dot._encoding)

    @property
    def nodes(self):
        return sorted(self._nodes, key=lambda x: x.title)

    def rules(self, depth=-1, index=True, viz_js=False):
        filenames, rules = set(), []
        rules.extend(self._rules(depth=depth, filenames=filenames, memo={}))
        for b, f in ((viz_js, self.site_viz), (index, self.site_index)):
            if b:
                rules.extend(list(update_filenames(f(self), filenames))[::-1])
        it = ((k, osp.join(*v).replace('\\', '/')) for k, v in reversed(rules))
        return collections.OrderedDict(it)

    def _rules(self, depth=-1, rule=(), filenames=None, memo=None):
        if self.foldername:
            rule += self.foldername,
        if filenames is None:
            filenames = set()
        if self.include_folders_as_filenames:
            filenames.update(v.foldername for k, v in self.items())
        if depth != 0:
            depth -= 1
            for folder, smap in self.items():
                folder_hash = hash(folder)
                if folder_hash in memo:
                    rules = memo[folder_hash]
                else:
                    memo[folder_hash] = rules = []
                    rules.extend(smap._rules(rule=rule, depth=depth, memo=memo))
                    for k, filename in update_filenames(folder, filenames):
                        rules.append((k, rule + filename))
                yield from rules

        for node in self._nodes:
            for k, filename in update_filenames(node, filenames):
                yield k, rule + filename

    def _add_obj(self, obj, workflow=False, folder=None, memo=None, **options):
        item = parent_func(obj)
        workflow &= not isinstance(item, NoSub)
        if workflow:
            item = self.get_sol_from(item)
            dsp, graph = item.dsp, item.workflow
        else:
            dsp = self.get_dsp_from(item)
            graph = dsp.dmap
        item_hash = hash(item)
        if item_hash in memo:
            folder, smap = memo[item_hash], None
            self[folder] = folder.sitemap
        else:
            memo[item_hash] = folder = self.site_folder(
                item, dsp, graph, obj, workflow=workflow, parent=folder,
                short_name=self.short_name, **options
            )
            folder.sitemap = smap = self[folder] = self.__class__()
            smap.short_name = self.short_name
        return smap, folder

    def add_items(self, item, workflow=False, depth=-1, folder=None, memo=None,
                  **options):
        opt = selector(self.options, self.__dict__, allow_miss=True)
        opt = combine_dicts(options, base=opt)
        if memo is None:
            memo = {}
        smap, folder = self._add_obj(
            item, workflow=workflow, folder=folder, memo=memo, **opt
        )
        if smap is None:
            return folder
        if depth > 0:
            depth -= 1
        site_node, append = self.site_node, smap._nodes.append
        add_items = functools.partial(
            smap.add_items, workflow=workflow, folder=folder, memo=memo, **opt
        )
        for node in itertools.chain(folder.nodes, folder.edges):
            links, node_id, node_title = node._links, node.node_id, node.title
            only_site_node = depth == 0 or (
                    node.type == 'data' and not node.attr.get('force_plot', 0)
            )
            for k, item in node.items():
                try:
                    if only_site_node:
                        raise ValueError
                    link = add_items(item, depth=depth, name=node_id, memo=memo)
                except ValueError:  # item is not a dsp object.
                    i = ''.join((node_title, k and '-' or '', k))
                    link = site_node(
                        folder, i, item, item, node_id, self.short_name
                    )
                    append(link)
                links[k] = link

        return folder

    @staticmethod
    def get_dsp_from(item):
        from ..sol import Solution
        from ...dispatcher import Dispatcher
        if isinstance(item, (Solution, SubDispatch)):
            return item.dsp
        elif isinstance(item, Dispatcher):
            return item
        raise ValueError('Type %s not supported.' % type(item).__name__)

    @staticmethod
    def get_sol_from(item):
        from ..sol import Solution
        from ...dispatcher import Dispatcher
        if isinstance(item, (Dispatcher, SubDispatch)):
            return item.solution
        elif isinstance(item, Solution):
            return item
        raise ValueError('Type %s not supported.' % type(item).__name__)

    def get_directory(self, directory):
        return directory or self.directory or tempfile.mkdtemp()

    def basic_app(self, root_path, mute=True, blueprint_name=None, **kwargs):
        import flask
        root_path = osp.abspath(self.get_directory(root_path))
        if blueprint_name is None:
            app = flask.Flask(root_path, root_path=root_path, **kwargs)
        else:
            app = flask.Blueprint(
                blueprint_name, root_path, root_path=root_path, **kwargs
            )
        app.before_request(functools.partial(before_request, mute))
        return app

    def app(self, root_path=None, depth=-1, index=True, mute=True, viz_js=False,
            executor='async', blueprint_name=None, **kw):
        rendered = {}
        app = self.basic_app(
            root_path, mute=mute, blueprint_name=blueprint_name, **kw
        )
        context = self.rules(depth=depth, index=index, viz_js=viz_js)

        if context:
            rules = {v: k[0] for k, v in context.items()}
            func = functools.partial(
                site_view, context, rendered, rules,
                next(iter(context.values())), viz=viz_js, executor=executor
            )
            app.add_url_rule('/<path:filepath>', 'default', view_func=func)
            app.add_url_rule('/<string:filepath>', 'default')
            app.add_url_rule('/', 'default', defaults={'filepath': ''})
        return app

    def site(self, root_path=None, depth=-1, index=True, view=False, run=False,
             **kw):
        site = Site(self, root_path=root_path, depth=depth, index=index, **kw)
        (run or view) and site.run()
        if view:
            # noinspection PyArgumentList
            self._view(site.url, format='html', quiet=True)

        return site

    def render(self, depth=-1, directory='static', view=False, index=True,
               viz=False, viz_js=False, executor='async'):

        directory = self.get_directory(directory)
        context = self.rules(depth=depth, index=index, viz_js=viz_js)
        rendered = {}
        for node, extra in context:
            if not extra:
                cached_view(
                    node, directory, context, rendered, viz=viz or viz_js,
                    executor=executor
                )
        for v in rendered.values():
            try:
                v.result()
            except Exception as ex:
                log.warning(ex)
        fpath = osp.join(directory, next((
            v for (i, j), v in context.items()
            if not isinstance(i, NoView) and j is None
        ), ''))
        if view:
            # noinspection PyArgumentList
            self._view(fpath, format=osp.splitext(fpath)[1][1:], quiet=True)
        return fpath


def _cached_view_task(
        node, directory, context, rend, viz, executor, **render_ctx):
    n_id = node.view_id
    rend = {k: v.result() for k, v in rend.items()}
    cnt = {(n_id, e): f for (n, e), f in context.items() if n == node}
    if rend and all(k in rend and osp.isfile(rend[k]) for k in cnt):
        for k, f in cnt.items():
            fpath = uncpath(osp.join(directory, f))
            os.makedirs(osp.dirname(fpath), exist_ok=True)
            parent, child = _compile_subs(rend[k], fpath)
            with open(fpath, 'w') as new_file:
                with open(rend[k]) as old_file:
                    for line in old_file:
                        new_file.write(child(parent(line)))
            rend[k] = fpath
    else:
        rend = node.view(
            osp.join(directory, context[(node, None)]), context=context,
            viz=viz, executor=executor, **render_ctx
        )
    return rend


def _set_rendered(results, fut, expected=None):
    try:
        error, res = None, fut.result()
    except Exception as ex:
        res, error = {}, ex
    for k, v in results.items():
        if k in res:
            v.set_result(res[k])
        else:
            v.set_exception(
                error or ValueError(f'Missing rendered result `{expected[k]}`.')
            )


def cached_view(node, directory, context, rendered, viz=False, executor='async',
                **render_ctx):
    n_id = node.view_id
    rend = {k: v for k, v in rendered.items() if k[0] == n_id}
    expected = {(n_id, e): f for (n, e), f in context.items() if n == node}
    res = {k: Future() for k in expected if k not in rendered}
    if res:
        rendered.update(res)
        PLOT_EXECUTORS.get_executor((executor, None)).thread(
            None, _cached_view_task, node, directory, context, rend, viz,
            executor, **render_ctx
        ).add_done_callback(functools.partial(
            _set_rendered, res, expected=expected
        ))
    rend.update(res)
    return rend


def _compile_subs(o, n):
    bn, dn, st = osp.basename, osp.dirname, osp.splitext
    return _sub(bn(dn(o)), bn(dn(n))), _sub(st(bn(o))[0], st(bn(n))[0])


def _sub(old, new):
    p, repl = r'(href\s*=\s*"[^"]*)(/%s)((.|/)[^"]*")' % old, r'\1/%s\3' % new
    return functools.partial(regex.compile(p, regex.IGNORECASE).sub, repl)



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build a form flask app from a dispatcher.
"""
import os
import secrets
import inspect
import os.path as osp

try:
    import flask_security
except ImportError:
    flask_security = None


class Config:
    DEBUG = True
    WTF_CSRF_CHECK_DEFAULT = False

    # reCAPTCHA configuration
    RECAPTCHA_PUBLIC_KEY = os.environ.get(
        'RECAPTCHA_PUBLIC_KEY', '6LcsgJglAAAAAMm7ilxkhBRevaCAuxlpefYZmxHU'
    )
    RECAPTCHA_PRIVATE_KEY = os.environ.get(
        'RECAPTCHA_PRIVATE_KEY', '6LcsgJglAAAAAAbR3aHm2qJS_c3XsGqmC9O816eH'
    )

    # Generate a nice key using secrets.token_urlsafe()
    SECRET_KEY = os.environ.get("SECRET_KEY", secrets.token_urlsafe())
    # Bcrypt is set as default SECURITY_PASSWORD_HASH, which requires a salt
    # Generate a good salt using: secrets.SystemRandom().getrandbits(128)
    SECURITY_PASSWORD_SALT = os.environ.get(
        "SECURITY_PASSWORD_SALT", secrets.SystemRandom().getrandbits(128)
    )
    SECURITY_URL_PREFIX = os.environ.get("SECURITY_URL_PREFIX", '/user')
    SECURITY_CONFIRMABLE = os.environ.get(
        "SECURITY_CONFIRMABLE", 'true'
    ).lower() == 'true'
    SECURITY_AUTO_LOGIN_AFTER_CONFIRM = os.environ.get(
        "SECURITY_AUTO_LOGIN_AFTER_CONFIRM", 'false'
    ).lower() == 'true'
    SECURITY_POST_CONFIRM_VIEW = os.environ.get(
        "SECURITY_POST_CONFIRM_VIEW", '/#login'
    )
    SECURITY_CONFIRM_ERROR_VIEW = os.environ.get(
        "SECURITY_CONFIRM_ERROR_VIEW", '/#login'
    )
    SECURITY_REGISTERABLE = os.environ.get(
        "SECURITY_REGISTERABLE", 'true'
    ).lower() == 'true'
    SECURITY_RECOVERABLE = os.environ.get(
        "SECURITY_RECOVERABLE", 'true'
    ).lower() == 'true'
    SECURITY_RESET_VIEW = os.environ.get("SECURITY_RESET_VIEW", '/#reset')
    SECURITY_RESET_ERROR_VIEW = os.environ.get(
        "SECURITY_RESET_ERROR_VIEW", '/#login'
    )
    SECURITY_REDIRECT_BEHAVIOR = os.environ.get(
        "SECURITY_REDIRECT_BEHAVIOR", 'spa'
    )

    SECURITY_TRACKABLE = os.environ.get(
        "SECURITY_TRACKABLE", 'true'
    ).lower() == 'true'

    MAIL_DEFAULT_SENDER = 'ProsaWeb <info@prosaweb.it>'
    MAIL_SERVER = os.environ.get('MAIL_HOST', 'smtps.aruba.it')
    MAIL_PORT = int(os.environ.get('MAIL_PORT', 465))
    MAIL_USE_SSL = True
    MAIL_USERNAME = os.environ.get('MAIL_USER', 'info@prosaweb.it')
    MAIL_PASSWORD = os.environ.get('MAIL_PASSWORD', 'prosaweb@123')

    SECURITY_I18N_DIRNAME = []
    if flask_security is not None:
        SECURITY_I18N_DIRNAME.append(osp.join(
            osp.dirname(inspect.getfile(flask_security)), 'translations'
        ))
    SECURITY_I18N_DIRNAME.extend((
        os.environ.get('SECURITY_I18N_DIRNAME', 'translations'),
        "translations"
    ))

    SCHEDULA_I18N_DIRNAME = [
        os.environ.get('SCHEDULA_I18N_DIRNAME', 'translations'),
        "translations"
    ]

    # have session and remember cookie be samesite (flask/flask_login)
    REMEMBER_COOKIE_SAMESITE = "strict"
    SESSION_COOKIE_SAMESITE = "strict"

    # Use an in-memory db
    SQLALCHEMY_DATABASE_URI = 'sqlite://'
    # As of Flask-SQLAlchemy 2.4.0 it is easy to pass in options directly to the
    # underlying engine. This option makes sure that DB connections from the
    # pool are still valid. Important for entire application since
    # many DBaaS options automatically close idle connections.
    SQLALCHEMY_ENGINE_OPTIONS = {"pool_pre_ping": True}
    SQLALCHEMY_TRACK_MODIFICATIONS = False



# !/usr/bin/env python
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

__all__ = ['__version__', '__updated__', '__title__', '__author__',
           '__license__', '__copyright__']

#: Authoritative project's PEP 440 version.
__version__ = version = "1.5.0"  # Also update README.rst

# Please UPDATE TIMESTAMP WHEN BUMPING VERSIONS AND BEFORE RELEASE.
#: Release date.
__updated__ = "2023-11-10 11:30:00"

__title__ = 'schedula'

__author__ = 'Vincenzo Arcidiacono <vinci1it2000@gmail.com>'

__license__ = 'EUPL, see LICENSE.txt'

__copyright__ = 'Copyright 2015-2023, Vincenzo Arcidiacono'



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2023, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides docutils nodes to plot dispatcher map and workflow.
"""
import html
import graphviz as gviz
from docutils import nodes as _nodes

ENGINES = ['dot', 'fdp', 'osage', 'neato', 'circo']
ENGINES += [k for k in sorted(gviz.ENGINES) if k not in ENGINES]


class _DspPlot(gviz.Digraph):
    _format = 'svg'

    def __init__(self, sitemap=None, *args, engines=None, **kwargs):
        super(_DspPlot, self).__init__(*args, **kwargs)
        self.sitemap = sitemap
        self.engines = [self.engine] + [
            k for k in engines or ENGINES if k != self.engine
        ]

    def render(self, *args, **kwargs):
        error = None
        for v in self.engines:
            self.engine = v
            try:
                return super(_DspPlot, self).render(*args, **kwargs)
            except Exception as ex:
                if error is None:
                    error = ex
        raise error


class _Table(_nodes.General, _nodes.Element):
    tagname = 'TABLE'

    def adds(self, *items):
        for item in items:
            # noinspection PyMethodFirstArgAssignment
            self += item
        return self


class _Tr(_Table):
    tagname = 'TR'

    def add(self, text, **attributes):
        # noinspection PyMethodFirstArgAssignment
        self += _Td(**attributes).add(text)
        return self


class _Td(_nodes.General, _nodes.Element):
    tagname = 'TD'

    def add(self, text):
        # noinspection PyMethodFirstArgAssignment
        self += _nodes.Text(html.escape(text).replace('\n', '<BR/>'))
        return self
